/*
* smd_export.ms
* Copyright (C) 2004 by Guillaume Plourde gplourde@centrenad.com

* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
Update: December 21 2004
This script is slow sorry. You can currently export Editable mesh with
Multi/Sub-object material and skin modifier or bones. Support for
Character studio 4 with physique is comming soon. I know a lot of people
work with editable poly now and support for that will be added in a short
time also. Vertex animation is also something I plan to finish. This script
UI is based on the Valve Software smd exporter for XSI. Send bug report or
code fix to gplourde@centrenad.com
*/

-- Close previous exporter dialog
try (DestroyDialog smd_export) catch ()

/*
------------------------------------------------------------------------
Globals
------------------------------------------------------------------------
*/

struct smdNode (
	name,
	parentIdx,
	obj
)

struct smdWeight (
	name,
	nodeID,
	weight
)

nodeArray = #()

/*
------------------------------------------------------------------------
Fix the float!
------------------------------------------------------------------------
*/

fn floatConvert n = (
	n = n as float
	i = (n as integer)
	f = abs (n - i)
	i = abs i
	if n < 0 then (	p = "-"	) else ( p = " " )
	f = (((1000000.0 * f) as integer) as string)
	fpos = (7 - f.count)
	flength = (7 - fpos)
	f = replace "000000" fpos flength f
	s = (p + (i as string) +"."+ (f as string))
	return s
)

/*
------------------------------------------------------------------------
Find the skin or physique modifier.
Support skin only, character studio physique is not working yet.
------------------------------------------------------------------------
*/

fn getskinmodifier o skinmod logFile = (
	for m = 1 to o.modifiers.count do (
		if classof o.modifiers[m] == Skin do (
			if skinmod == 0 do (
				skinmod = o.modifiers[m]
				max modify mode
				modPanel.SetCurrentObject skinmod
			)
			exit
		)
	)
	return skinmod
)

/*
------------------------------------------------------------------------
Get the number of weights and create a smdWeight for each.
------------------------------------------------------------------------
*/

fn getWeights o skinmod v logFile = (
	warray = #()
	if skinmod == 0 then (
		warray = #(0)
		return warray
	) else (
		try (
		wNum = skinOps.GetVertexWeightCount skinmod v
		if wNum > 5 do (format "ERROR: Vertex % as too many weight. Weight limit is 5\n" v to:logFile;)
		for w = 1 to wNum do (
			vert = skinOps.GetVertexWeight skinmod v w
			b = skinOps.GetVertexWeightBoneID skinmod v w
			n = skinOps.GetBoneName skinmod b 0
			append warray (smdWeight n b (floatConvert vert))
		)
		) catch (
		)
		return warray
	)
)

/*
------------------------------------------------------------------------
Make a nice table of strings for the output
------------------------------------------------------------------------
*/
fn makeWeightTable warray = (
	weights = ""
	try (
		for i in warray do (
			weights += (i.nodeID as string +" "+ i.weight as string +" ")
		)
	) catch ()
	return weights
)

/*
------------------------------------------------------------------------
Write nodes to the smd file
------------------------------------------------------------------------
*/

fn getNodes fs o = (
	select o
	skinmod = 0

	-- put mesh geometry before bones
	if classof o == Editable_mesh do (
		-- check scene geometry and add them to the node list
		b = smdNode name:o.name obj:o parentIdx:0
		append nodeArray b
		
		-- Check for skin modifier
		if skinmod == 0 do (
			skinmod = getskinmodifier o skinmod logfile
		)
		
		if skinmod != 0 do (
			nbones = skinOps.GetNumberBones skinmod
			for i = 1 to nbones do (
				n = skinOps.GetBoneName skinmod i 0
				b = smdNode name:n parentIdx:0
				append nodeArray b
			)
		)	
	)
)

fn formatNodes fs = (
	-- Check each node for a parent and send the parent index to the current node
	for i = 1 to nodeArray.count do (
		t = 0
		
		-- We match the name to a node in the scene
		for n = 1 to geometry.count do (
			if nodeArray[i].name == geometry[n].name do (
				nodeArray[i].obj = geometry[n]
			)
		)
	
		if nodeArray[i].obj.parent != undefined do (
			t = nodeArray[i].obj.parent.name
		)
		
		-- Then we match the parent name with the name of every node in the list
		for n = 1 to nodeArray.count do (
			if t == 0 do (
				exit
			)
			if nodeArray[n].name == t do (
				nodeArray[i].parentIdx = n
				exit
			)
		)
		
		-- Write node index, name and parent index to the smd file
		format "% \"%\" %\n" (i - 1) nodeArray[i].name (nodeArray[i].parentIdx - 1) to:fs
	)
)

/*
------------------------------------------------------------------------
Write skeleton to the smd file
Bones only, Character studio 4.3 support comming soon
------------------------------------------------------------------------
*/
fn formatSkeleton fs a = (
	format "skeleton\n" to:fs
	
	t = 0
	tend = filterString (animationRange.end as string) "f"
	tend = tend[1]
	
	while t <= (tend as integer) do (
		try (
		slidertime = t
		format "time %\n" t to:fs;
		for i = 1 to nodeArray.count do (
			-- do we have a parent?
			if nodeArray[i].parentIdx != 0 then (
				n = nodeArray[i].obj
				
				if classof n == Biped_Object then (
	
				) else (
					r = ((n.transform.rotationPart * (inverse n.parent.transform.rotationPart)) as EulerAngles)
				)
				
				rotInRadX = (degToRad r.x)
				rotInRadY = (degToRad r.y)
				rotInRadZ = (degToRad r.z)
				
				in coordsys parent format "% % % % % % %\n" (i - 1) (floatConvert nodeArray[i].obj.pos.x) (floatConvert nodeArray[i].obj.pos.y) (floatConvert nodeArray[i].obj.pos.z) (floatConvert rotInRadX) (floatConvert rotInRadY) (floatConvert rotInRadZ) to:fs		

			) else (
				n = nodeArray[i].obj
				
				if classof n == Biped_Object then (
	
				) else (
					r = in coordsys world (n.transform.rotationPart as EulerAngles)
				)
				
				rotInRadX = (degToRad r.x)
				rotInRadY = (degToRad r.y)
				rotInRadZ = (degToRad r.z)
		
				format "% % % % % % %\n" (i - 1) (floatConvert nodeArray[i].obj.pos.x) (floatConvert nodeArray[i].obj.pos.y) (floatConvert nodeArray[i].obj.pos.z) (floatConvert rotInRadX) (floatConvert rotInRadY) (floatConvert rotInRadZ) to:fs		
			)
		)
		) catch (
		)
		t += 1
		if a == 0 do (exit)
	)
	
	format "end\n" to:fs
)

/*
------------------------------------------------------------------------
Write VTA vertex animation
------------------------------------------------------------------------
*/
fn formatVertAnim fs = (

	format "vertexanimation\n" to:fs
	
	t = 0
	tend = filterString (animationRange.end as string) "f"
	tend = tend[1]
	prevMesh = 0
	while t <= (tend as integer) do (
		slidertime = t
		format "time %\n" t to:fs;
		for i = 1 to nodeArray.count do (
			o = nodeArray[i].obj
			if classof nodeArray[i].obj == Editable_mesh do (
				n = meshop.getNumVerts nodeArray[i].obj
				for v = 1 to n do (
					p = meshop.getVert nodeArray[i].obj v
					norm = getNormal nodeArray[i].obj v
					if (prevMesh != 0) then (
						pp = meshop.getVert prevMesh v
						if pp != p do (
							format "% % % % % % %\n" (v - 1) (floatConvert p.z) (floatConvert p.y) (floatConvert p.z) (floatConvert norm.x) (floatConvert norm.y) (floatConvert norm.z) to:fs
						)
					) else (
						format "% % % % % % %\n" (v - 1) (floatConvert p.z) (floatConvert p.y) (floatConvert p.z) (floatConvert norm.x) (floatConvert norm.y) (floatConvert norm.z) to:fs
					)
					prevMesh = snapshotAsMesh o
				)
			)
		)
		t += 1
		prevMesh = snapshotAsMesh nodeArray[1].obj
	)
	
	format "end\n" to:fs
)

/*
------------------------------------------------------------------------
Write mesh to the smd file
------------------------------------------------------------------------
*/

fn formatMesh fs o logFile = (
	select o
	oId = 0
	for i = 1 to nodeArray.count do (
		if nodeArray[i].name == o.name do (
			oID = i
		)
	)
	if classof o == Editable_mesh do (
		skinmod = 0
		-- Check if the mesh as a multisub material.
		if (classof o.material != Multimaterial) do (
			format ("WARNING: Mesh with no Multi/Sub-Object material.\n") to:logFile
		)
		
		for f = 1 to (meshop.GetNumFaces o) do (
			
			-- Get the vertices
			tri = getFace o f
			p1 = getVert o tri.x
			p2 = getVert o tri.y
			p3 = getVert o tri.z
			pv1x = p1.x
			pv1y = p1.y
			pv1z = p1.z
			pv2x = p2.x
			pv2y = p2.y
			pv2z = p2.z 
			pv3x = p3.x
			pv3y = p3.y
			pv3z = p3.z
			
			-- Get the normal
			faceNormal = in coordsys world meshop.getFaceRNormals o f
			n1 = faceNormal[1]
			n2 = faceNormal[2]
			n3 = faceNormal[3]
			
			-- Get Texture Coords
			try (
				tvface = getTVFace o f
				t1 = getTVert o tvface.x
				t2 = getTVert o tvface.y
				t3 = getTVert o tvface.z
			) catch (
				-- Add error log
				format "WARNING: Face with no UV mapping.\n" to:logFile
				t1 = [0,0,0]
				t2 = [0,0,0]
				t3 = [0,0,0]
			)
			
			-- Get Material ID
			try (
				m = getFaceMatID o f
				if (o.material[m] == undefined) do (
					format "WARNING: Face MatID with no matching material.\n" to:logFile
				)
				m = filenamefrompath o.material[m].diffusemap.filename
				-- todo, undefined?
			) catch (
				m = "ERROR.TGA"
				format "WARNING: Texture file not found.\n" to:logFile
			)
			
			-- Look for a skin modifier
			if skinmod == 0 do (
				skinmod = getSkinmodifier o skinmod logFile
			)
			
			-- Check skinmod for modifier, if skinmod  is 0 then skip all that
			if skinmod != 0 then (
				warray_v1 = getWeights o skinmod tri.x logFile
				warray_v2 = getWeights o skinmod tri.y logFile
				warray_v3 = getWeights o skinmod tri.z logFile
	
				weights_v1 = makeWeightTable warray_v1
				weights_v2 = makeWeightTable warray_v2
				weights_v3 = makeWeightTable warray_v3
				
				skin_v1 = (warray_v1.count as string +" "+ weights_v1) 
				skin_v2 = (warray_v2.count as string +" "+ weights_v2)
				skin_v3 = (warray_v3.count as string +" "+ weights_v3)
			) else (
				-- If no skin modifier is found then leave blank. Vertex will be assign 100% to the mesh.
				skin_v1 = ""
				skin_v2 = ""
				skin_v3 = ""			
			)
			
			try (
				test = n1.x
				format "%\n" m to:fs;			
				format "% % % % % % % % % %\n" (oId - 1) (floatConvert pv1x) (floatConvert pv1y) (floatConvert pv1z) (floatConvert n1.x) (floatConvert n1.y) (floatConvert n1.z) (floatConvert t1.x) (floatConvert t1.y) skin_v1 to:fs;
				format "% % % % % % % % % %\n" (oId - 1) (floatConvert pv2x) (floatConvert pv2y) (floatConvert pv2z) (floatConvert n2.x) (floatConvert n2.y) (floatConvert n2.z) (floatConvert t2.x) (floatConvert t2.y) skin_v2 to:fs;
				format "% % % % % % % % % %\n" (oId - 1) (floatConvert pv3x) (floatConvert pv3y) (floatConvert pv3z) (floatConvert n3.x) (floatConvert n3.y) (floatConvert n3.z) (floatConvert t3.x) (floatConvert t3.y) skin_v3 to:fs;			
			) catch (
				-- jit
			)
		)
	)
)

/*
------------------------------------------------------------------------
Run this when everything is done or if the script catch an error.
------------------------------------------------------------------------
*/

fn closescript fs showlog logFile = (
	close fs
	close logFile
	if showlog == 1 do (
		shelllaunch "notepad" ((getdir #scripts) + "\smdexport.log")
	)
)

rollout smd_export "Source .SMD Exporter"
(
	group "File" (
		dropdownList ddl_fileType "File Type" items:#("Model (.SMD)", "Skeletal Animtation (.SMD)", "Vertex Animation (.VTA)")
	)
	group "Options" (
		checkbox showlog "Show exporter log" checked:false
	)
	label lb1 "by Guillaume Plourde gplourde@centrenad.com" align:#right
	button acpt "Ok" pos:[104,190] width:90
	on acpt pressed do (
		meshToExport = #()
		for i in selection do (
			if classof i == Editable_mesh do (
				append meshToExport i
			)
		)	
		try (
			logFile = openFile (getdir #scripts + "\smdexport.log") mode:"w"
		) catch (messagebox "Couldn't open file for writing.")
		if (ddl_fileType.selection == 1) then (
			try (
				f = getSaveFileName caption:"Save as" types:"Valve (*.smd)|*.smd"
			) catch (messagebox "Couldn't open file for writing.")
			if f != undefined do (
				L = 0
				if showlog.checked == true do (
					L = 1
				)
				fs = openFile f mode:"w"		
				format "version 1\n" to:fs
				for i in meshToExport do (getNodes fs i)
				format "nodes\n" to:fs
				formatNodes fs
				format "end\n" to:fs
				formatSkeleton fs 0
				format "triangles\n" to:fs
				for j in meshToExport do (formatMesh fs j logFile)
				format "end\n" to:fs
				closeScript fs L logFile
				DestroyDialog smd_export
			)
		)
		else if (ddl_fileType.selection == 2) then (
			try (
				f = getSaveFileName caption:"Save as" types:"Valve (*.smd)|*.smd"
			) catch (messagebox "Couldn't open file for writing.")
			if f != undefined do (
				L = 0
				if showlog.checked == true do (
					L = 1
				)
				fs = openFile f mode:"w"		
				format "version 1\n" to:fs
				for i in meshToExport do (getNodes fs i)
				format "nodes\n" to:fs
				formatNodes fs
				format "end\n" to:fs
				formatSkeleton fs 1
				closeScript fs L logFile
				DestroyDialog smd_export
			)
		)
		else if (ddl_fileType.selection == 3) then (
			try (
				f = getSaveFileName caption:"Save as" types:"Valve (*.vta)|*.vta"
			) catch (messagebox "Couldn't open file for writing.")
			if f != undefined do (
				L = 0
				if showlog.checked == true do (
					L = 1
				)
				fs = openFile f mode:"w"		
				format "version 1\n" to:fs
				formatNodes fs
				formatSkeleton fs 0
				formatVertAnim fs
				closeScript fs L logFile
				DestroyDialog smd_export
			)
		)
	)
	
	button quit "Cancel" pos:[200,190] width:90
	on quit pressed do (
		DestroyDialog smd_export
	)
)
createdialog smd_export 300 218